#
#
#  gen_config.perl
#
#    Copyright (c) 1994, 1995, 1996 by Merit Network, Inc.
#    All rights reserved.
#
#    Generates gated configuration files.
#
#    Implementation notes:
#
#    * The BSD Unix version of perl is different from Sun Unix.
#      To keep the code portable, the following must be done:
#      - Any local variable declarations must be put in ().
#      - End of input line must be checked before chopping newline;
#        BSDI has \n at end of input line, Sun does not.
#
#
#  Author:  Jeff Jackson
#
#  Revision History:
#
#    01-26-95   JWJ   First version finished.
#    02-23-95   JWJ   Added -m option, default file path
#    03-01-95   JWJ   Added IP_STATIC_OPEN, IP_STATIC_CLOSE
#                       (to work around gated parser bug)
#    03-02-95   JWJ   Added socket specification
#    03-03-95   JWJ   Minor environment variable name changes
#    03-07-95   JWJ   Minor bug fix in find_machine (for < to for <=)
#    03-23-95   JWJ   Fixed bug - rdi of internal neighbors should
#                     match rdi of host machine of config file
#    03-30-95   JWJ   Added name of this file to config file headers.
#    04-27-95   JWJ   Added ROUTE-SERVER, MULTI-EXIT-DISC options.
#    05-04-95   JWJ   Added ip_gw switch.
#    05-08-95   JWJ   Bug fixes to route_server and multi_exit code.
#                     Changed ip_gw code to be machine-specific.
#    05-10-95   JWJ   Added code to overwrite machine's default RDI
#                     with that from another machine.  Removed the
#                     "bug fix" of 3-23-95 (the solution was insufficient
#                     and is not trivial).  New solution is to have the
#                     user specify the RDI in the .cfg file.
#    06-07-95   JWJ   Added "net" keyword in neighbor statement before
#                     osi address (now required).
#    08-04-95   JWJ   Added env var checking.
#                     Added "traceoptions" var processing.
#                     Changed machine map field parsing (\t to \s).
#                     and fixed portability bug on chop (see below).
#                     Added different gateway (gw) for default route.
#    08-18-95   JWJ   Changed to always put in IP "default" route.
#                     Changed some terminology to avoid confusion.
#                     (ex: IP_DEFAULT_ROUTES to IP_COMMON_STATIC_ROUTES)
#    08-21-95   JWJ   Removed IP_STATIC_OPEN and IP_STATIC_CLOSE
#                     (change - both IP and ISO now use "static" area).
#                     Changed "ip_gw" keyword to "gateway" (ISO routes now use it).
#                     Added first pass of "export" statement handling.
#    08-24-95   JWJ   Eliminated duplicate RDI in import/export processing.
#    09-07-95   JWJ   Put default route name and common static route info
#                     in machine map file.  Added keywords for processing.
#                     Added global IP and OSI prefixes to machine map file.
#    09-28-95   JWJ   Added import and export processing using template files
#                     and setvar handling.
#    09-29-95   JWJ   Updated variable processing (unique to a test set).
#    10-05-95   JWJ   Added overview and usage messages (-h option or error).
#                     Added code to avoid results file generation (-E option).
#                     Added code to avoid export loop code (-L option).
#                     Fixed bug in internal/external neighbor processing
#                      (two separators needed).
#    10-11-95   JWJ   Put in general variable substitution.
#                     Added a more general and helpful error handling routine.
#    10-16-95   JWJ   Fixed user variable substitution in import/export code.
#    10-17-95   JWJ   Put in new RIB generation code (1 template per RIB).
#                     Added -W option (suppress warning messages).
#    10-24-95   JWJ   Added specific RDI and RDPATH processing.
#    10-25-95   JWJ   Added &rdi and &net built-in functions.
#    10-27-95   JWJ   Modified expected result file generation to include
#                     only the RIBs in the "ribs-supported" peer clauses.
#    10-30-95   JWJ   Removed the "$RDI" predefined variable (for stepping
#                     through RDI list without looping).  More efficient now
#                     to use user-defined vars + built-in functions.
#    11-01-95   JWJ   Added code to put in "." when generating OSI addresses
#                     if length of last prefix + index is longer than 4.
#    11-28-95   JWJ   Replaced route range selection (none, some, all)
#                     with route set processing.
#    12-5-95    JWJ   Removed predefines "multi-exit" and "route-server".
#                     Changed predefined vars to have $$ prefix.
#    12-20-95   JWJ   Added $$INCLUDE directive processing.
#    12-21-95   JWJ   Added OSPF processing (areas/interfaces).
#                     Removed process_option (obsolete).
#    01-04-96   JWJ   Updated to match new language configuration, based on
#                     sequence/routing domain/machine/protocol/peer|interface.
#                     Removed rdi keyword and set_rdi routine.  RDI is now
#                     set by "routing_domain" keyword.
#    01-08-96   JWJ   Removed "rdpath" processing (identical to single RDI).
#                     Changed "import/export_rdi_loop" to "import/export_all".
#                     Added the capability to put restrictions on the
#                     generation of sequence/test set combinations.
#    01-24-96   JWJ   Added user variable processing in the route set file.
#                     Added env variable "gated_template_dir" to remove the
#                     necessity of putting hard-coded path info in .cfg files.
#                     Did the same for $$INCLUDE directives.
#    01-25-96   JWJ   Modified route set code to specify route set ID for
#                     each machine in a test set (to prevent having to put
#                     in redundant route sets).
#    01-31-96   JWJ   Added "main_template" keyword (overrides env variable).
#                     Modified template lookup code to use search path rather
#                     than automatically looking in template directory.  (See
#                     revision history for 01-24-96.)
#                     Added "as_number" to machine map file (for BGP).
#    02-01-96   JWJ   Added "rib_template" command (to match up with
#                     handling of peer/area/interface templates).
#                     Removed restriction of no variable substitution in
#                     main template (other templates do substitute).
#                     Changed some predefined variables from directives
#                     to simple substitutions.
#    02-07-96   JWJ   Added IS-IS protocol support.
#    02-12-96   JWJ   Added "-t <test name>" option.  This changes the working
#                     directory to $gated_test_dir/<test name> and looks for
#                     "<test name>.cfg" as file to use.  (A shortcut).
#                     Added "-I" (suppress informational messages) option.
#    02-13-96   JWJ   Changed default route info from env variables to
#                     reading from route set file, as you can now have more
#                     than one default route.
#                     Removed "gateway" and "gateway all" commands (no longer
#                     necessary - using route set file).  Also removed ip_gw
#                     and osi_gw variables.
#                     Removed "internal_static_osi/ip_addr" variables and
#                     "default_ip_route" - read from machine map file (no
#                     longer used).
#    02-26-96   JWJ   Modified to allow multiple default route sets.
#    02-27-96   JWJ   Changed "default_routes" keyword to "route_set default".
#                     Changes to allow "all" to be used with interface keyword.
#                     Added the capability of setting user variables at the
#                     peer/interface level (set for one peer, different for
#                     the next in the same sequence).
#    03-07-96   JWJ   Changed "no route set for machine X in a test set"
#                     from error to warning.
#                     Fixed bug in result file creation which stepped on the
#                     default routes before they were written out to file.
#                     Added "null" as route set keyword.
#                     Changed "get_name" to have reference parameter so a
#                     name value of "0" won't cause an error in get_name.
#    03-11-96   JWJ   Changed several subroutine names to reflect idea of
#                     .cfg file command = "set", writing to file = "process".
#                     $$INCLUDE file name can now be a user variable.
#    03-12-96   JWJ   Changed user variable search to allow "$variable;"
#                     (no need for space between variable and semicolon).
#    03-27-96   JWJ   Moved BGP "allow_list" var to predefined var list.
#    04-01-96   JWJ   Renamed some global variables (g200->g160, new 201,202).
#                     Clarified the explanation of "export loop code."
#    04-08-96   JWJ   Made some major modifications to properly handle BGP
#                     external groups (one group for each AS).  Changes:
#                     Added the "group_template" command.  Replaced the
#                     $$INTERNAL_NEIGHBORS and $$EXTERNAL_NEIGHBORS directives
#                     with $$GROUPS and $$PEER_LOOP.
#    04-09-96   JWJ   Fixed a bug causing $$LOOP lines to be printed if no
#                     lines were generated as a result of the loop directive
#                     (added a "next" to the end of the code block).
#                     Added capability to specify group/peer/area/interface
#                     templates based on protocol.  This is necessary for .cfg
#                     files which use both BGP and IDRP, or OSPF and RIP.
#    04-11-96   JWJ   Fixed bug in result file creation ("!=" used instead of
#                     "ne" in string comparison).
#                     Changed $$GROUPS directive to $$BGP_GROUPS and
#                     $$IDRP_GROUPS (necessary in order to set proper
#                     protocol when generating peers).
#    04-12-96   JWJ   Fixed OSPF backbone statement; removed "backbone" command
#                     and put in "set_area" command instead.  This is to allow
#                     proper setting of a template with the backbone.
#    04-17-96   JWJ   Added BGP "igp", "routing", and "test" peer commands.
#    04-18-96   JWJ   Removed restriction on "protocol" command scope.  It
#                     can now be specified at any scope.
#                     Added the "mesh <peer type>=<machine list>" command.
#    04-23-96   JWJ   Fixed bug generating external mesh (was not setting the
#                     RDI to the current machine in the mesh).
#    04-24-96   JWJ   Put general predefined variables in their own routine,
#                     and added calls to it wherever user variables are also
#                     processed.  Removed the global "predef_var_list" as there
#                     is no real need for it.  Created a new global variable
#                     for holding proto-sock.
#    04-25-96   JWJ   Clarified error statement in mesh code.
#                     Rearranged CFG language commands in alphabetical order (to help
#                     with updating documentation).
#
#
#  Environment variables looked for:
#
#      gated_script_dir      Directory containing gated scripts
#      gated_template_dir    Directory containing gated templates
#      gated_machine_map     Machine map file
#      gated_template_file   File used as gated.conf template
#      gated_mainlog_dir     Directory of gated log file
#      gated_log_file_name   Name of gated log file
#
#  Important global variables:
#
#      g001_mach_idx         Current machine being processed
#      g002_curr_sequence    Current sequence
#      g003_curr_test_set    Current test set
#      g004_curr_peers       Current peers of current machine
#      g005_process_scope    Current process level (global, sequence, ...)
#      g006_curr_protocol    Current protocol
#      g007_curr_area        Current OSPF area
#      g011_curr_rdi         Current routing domain
#      g012_curr_as          Current AS (autonomous system) number
#      g013_curr_peer_intf   Current peer or interface (for setvars)
#      g021_curr_test_name   Current test name
#      g031_proto_sock       Current protocol socket
#
#      g100_data_list        List of machines/peers/interfaces in a sequence
#      g110_machine_list     Unique list of machines in a sequence
#      g112_area_template_list  List of OSPF area templates in a sequence
#      g120_import_stmt_list    List of import statements in a test set
#      g130_export_stmt_list    List of export statements in a test set
#      g140_route_set_id_list   List of machine/route set ID pairs by test set
#      g150_default_route_list  List of default routes (from route set file)
#      g160_generation_list  List of restrictions on generating sequence/
#                            test set combinations
#      g170_template_list    Master list of protocol-based templates
#
#      g201_opt_gen_res_files   Do/do not generate expected results files
#      g202_opt_gen_loop_code   Do/do not generate export loop code
#
#      Template directives processed:
#	$$HEADER
#	$$TRACEOPTIONS
#	$$RDI_NUMBER
#       $$AS_NUMBER
#	$$NET
#	$$CAPACITY
#	$$RIB_LOOP
#       $$PEER_LOOP
#	$$AREAS
#	$$OSPF_INTERFACES
#	$$RIP_INTERFACES
#	$$BGP_GROUPS
#	$$IDRP_GROUPS
#	$$STATIC_ROUTES
#	$$IMPORT_STATEMENTS
#	$$EXPORT_STATEMENTS
#

$PROGRAM_NAME = $0;
$VERSION      = "3.77";
$VDATE        = "April 25, 1996";

@arr = split( "/", $PROGRAM_NAME);
$BASENAME = $arr[$#arr];

$TRUE  = 1;
$FALSE = 0;

require "ctime.pl";
$date = &ctime(time);
chop( $date );

#
#  Option processing.
#    To see the full list of options and descriptions, see the
#    "print_usage" subroutine.
#
require "getopts.pl";
&Getopts('m:t:EhILW');   #sets $opt_m, $opt_t as arguments to -m, -t

#
#  Now make sure a test configuration file was specified.
#
if ( $opt_h )
{
   &print_overview;
   &print_usage;
   exit;
}

if ( ($#ARGV == -1) && (! $opt_t) )
{
   &print_usage;
   exit;
}

#
#  Suppress information messages by redirecting standard output to null.
#  (Errors go to STDERR, so "die" commands will continue to print.)
#
if ( $opt_I )
{
   open( NULL_OUT, ">/dev/null" )
      || die ' *** ERROR - Cannot open "/dev/null" for output\n';
   select( NULL_OUT );
}

#
#  Print header
#
print "\n";
print "              GATED Configuration File Generator\n";
print "               Version $VERSION - $VDATE\n";
print "          Copyright 1994, 1995, 1996 Merit Network, Inc.\n";
print "\n";

#
#  Default path and file names - using env vars
#  Exit if not declared.
#
$gated_script_dir   = $ENV{'gated_script_dir'};
$gated_test_dir     = $ENV{'gated_test_dir'};
$gated_template_dir = $ENV{'gated_template_dir'};
$machine_map_file   = $ENV{'gated_machine_map'};
$route_set_file     = $ENV{'gated_rte_set_file'};
$main_template_file = $ENV{'gated_template_file'};
$log_file_dir       = $ENV{'gated_mainlog_dir'};
$log_file_name      = $ENV{'gated_log_file_name'};

$success = 1;
if ( $gated_script_dir eq "" )
{
   print "   *** ERROR - \$gated_script_dir not set\n";
   $success = 0;
}
if ( $gated_test_dir eq "" )
{
   print "   *** ERROR - \$gated_test_dir not set\n";
   $success = 0;
}
if ( $gated_template_dir eq "" )
{
   print "   *** ERROR - \$gated_template_dir not set\n";
   $success = 0;
}
if ( $machine_map_file eq "" )
{
   print "   *** ERROR - \$gated_machine_map not set\n";
   $success = 0;
}
if ( $route_set_file eq "" )
{
   print "   *** ERROR - \$gated_route_set_file not set\n";
   $success = 0;
}
if ( $main_template_file eq "" )
{
   print "   *** ERROR - \$gated_template_file not set\n";
   $success = 0;
}
if ( $log_file_dir eq "" )
{
   print "   *** ERROR - \$gated_mainlog_dir not set\n";
   $success = 0;
}
if ( $log_file_name eq "" )
{
   print "   *** ERROR - \$gated_log_file_name not set\n";
   $success = 0;
}

if ( $success == 0 )
{
   die "   Edit and run the file 'main_setup.csh' to set up the necessary environment variables.\n";
}

print "     gated script directory = \"$gated_script_dir\"\n\n";

#
#  Process command line options.
#
if ( $opt_m ne "" ) {
   print "\nGenerating files only for machine $opt_m\n\n"; }

#
#  The .cfg file is either generated from the test name if -t is used,
#  or from an explicit name on the command line.
#
$config_file = ( $opt_t ne "" ? "$opt_t".".cfg" : $ARGV[0] );
if ( $opt_t ne "" )
{
   $work_dir = "$gated_test_dir/$opt_t";
   chdir ( $work_dir )
      || die " *** ERROR - Cannot change to directory \"$work_dir\"\n";
   print "Changing to $work_dir ...\n";
}

$print_warnings = ( $opt_W ) ? OFF : ON;
if ( ! $print_warnings ) {
   print "Suppressing warning messages\n"; }

#
#  Structure of g120_import_stmt_list and g130_export_stmt_list:
#     $list{order,test_set,protocol,type,index} = template
#
#     order       In gated, import/export statement order is significant.
#                 This preserves the order specified in the .cfg files.
#     test_set    Name of the test set
#     protocol    Protocol of this import/export statement
#     type        SINGLE, LOOP
#                 LOOP adds a sub-statement for every peer of the host machine
#     index       To allow more than one import/export statement per RDI
#                 to be stored in this array
#     template    Import/export template to use when processing this entry
#
%g120_import_stmt_list  = ();
%g130_export_stmt_list  = ();
%g140_route_set_id_list = ();
$import_order = 10000;  # for keeping array sorted in order of stuffing
$export_order = 10000;  # for keeping array sorted in order of stuffing

$ALL = "#all#";
$default_rib = "0";

@global_var_list   = ();
@sequence_var_list = ();
@machine_var_list  = ();
@test_set_var_list = ();
@entire_ribs_list  = ();
@peer_ribs_list    = ();

$num_cfg_files = 0;
$num_res_files = 0;
$num_sequences = 0;
$gl_exp_count  = 0;

$g031_proto_sock = idrp;

$g201_opt_gen_res_files = on;
$g202_opt_gen_loop_code = on;

@builtin_list = ( "\&as", "\&rdi", "\&net", "\&ip" );

@supported_protocol_list = ( bgp, idrp, isis, ospf, rip );

#  Turn off generation of results files if specified
if ( $opt_E )
{
   $g201_opt_gen_res_files = off;
   print "Expected results files will NOT be generated\n";
}

#  Turn off export loop code if specified.
#
#  "Export loop" means to export to a remote RDI routes which have been
#  imported from the host node's RDI.
#  Example:  YYY is the RDI of the host machine.
#     export proto idrp rdi XXX
#     {
#          proto idrp rdi YYY
#     }
#
if ( $opt_L )
{
   $g202_opt_gen_loop_code = off;
   print "Export loop code will NOT be generated\n";
}

#
#  Beginning scope = global scope.
#
$g005_process_scope = GLOBAL;

#
#  Read in information on machines.
#
&read_machine_map();

#
#  Look through the configuration file, processing lines which
#  define a sequence, ignoring comments and printing error otherwise.
#
open( TEST, $config_file )
   || die " *** ERROR - Cannot open test configuration file \"$config_file\": $!\n";

$line_number = 0;
while ( <TEST> )
{
#  print" ** LINE:  $_";
   TEST_SWITCH:
   {
      $line_number++;

      #  Establish the current OSPF area.
      /^\s*area[\s\n]+/ && ( &set_area, last TEST_SWITCH );

      #  Establish the current OSPF area template filename.
      /^\s*area_template[\s\n]+/ && ( &set_template(area_template), last TEST_SWITCH );

      #  Assign a list of BGP "test" neighbors of current machine.
      /^\s*bgp_test[\s\n]+/  && ( &set_peers(test), last TEST_SWITCH );

      #  Synonym of "sequence" (identical).
      /^\s*configuration[\s\n]+/ && ( &begin_sequence, last TEST_SWITCH );

      #  End of sequence.  Create configuration files.
      /^\s*end[\s\n]+/ && ( &end_block, last TEST_SWITCH );

      #  Specify a specific template for exporting.
      /^\s*export[\s\n]+/ && ( &set_ie_template(export), last TEST_SWITCH );

      #  Specify that a loop is used for export and get template to use.
      /^\s*export_all[\s\n]+/ && ( &set_ie_template(export_all), last TEST_SWITCH );

      #  Assign a list of external neighbors of current machine.
      /^\s*external[\s\n]+/  && ( &set_peers(external), last TEST_SWITCH );

      #  Restriction entry on generating sequence/test set combinations.
      /^\s*generate[\s\n]+/ && ( &add_generate_entry, last TEST_SWITCH );

      #  Establish the current IDRP/BGP group template filename.
      /^\s*group_template[\s\n]+/ && ( &set_template(group_template), last TEST_SWITCH );

      #  Assign a list of BGP "igp" neighbors of current machine.
      /^\s*igp[\s\n]+/  && ( &set_peers(igp), last TEST_SWITCH );

      #  Specify a specific template for importing.
      /^\s*import[\s\n]+/ && ( &set_ie_template(import), last TEST_SWITCH );

      #  Specify that a loop is used for import and get template to use.
      /^\s*import_all[\s\n]+/ && ( &set_ie_template(import_all), last TEST_SWITCH );

      #  Add a list of OSPF or RIP interfaces to current area or machine.
      /^\s*interface[\s\n]+/  && ( &add_interfaces, last TEST_SWITCH );

      #  Assign a list of internal neighbors of current machine.
      /^\s*internal[\s\n]+/  && ( &set_peers(internal), last TEST_SWITCH );

      #  Establish the current OSPF/RIP interface template filename.
      /^\s*intf_template[\s\n]+/ && ( &set_template(intf_template), last TEST_SWITCH );

      #  Set the current machine in the current test sequence.
      /^\s*machine[\s\n]+/ && ( &begin_machine(), last TEST_SWITCH );

      #  Set the main template filename (override environment variable).
      /^\s*main_template[\s\n]+/ && ( &set_template(main_template), last TEST_SWITCH );

      #  Define a full mesh of machines peering to each other.
      /^\s*mesh[\s\n]+/  && ( &set_full_mesh, last TEST_SWITCH );

      #  Establish the current IDRP/BGP peer template filename.
      /^\s*peer_template[\s\n]+/ && ( &set_template(peer_template), last TEST_SWITCH );

      #  Establish the current protocol.
      /^\s*protocol[\s\n]+/ && ( &set_protocol, last TEST_SWITCH );

      #  Assign a "rib" statement, in global or machine scope.
      /^\s*rib[\s\n]+/ && ( &set_rib, last TEST_SWITCH );

      #  Establish the current RIB template filename.
      /^\s*rib_template[\s\n]+/ && ( &set_template(rib_template), last TEST_SWITCH );

      #  Assign a "ribs-supported" statement of a peer.
      /^\s*ribs-supported[\s\n]+/ && ( &set_ribs_supported, last TEST_SWITCH );

      #  Establish the static routes to be used in this test set.
      /^\s*route_set[\s\n]+/ && ( &set_route_set, last TEST_SWITCH );

      #  Assign a list of BGP "routing" neighbors of current machine.
      /^\s*routing[\s\n]+/  && ( &set_peers(routing), last TEST_SWITCH );

      #  Establish the current routing domain.
      /^\s*routing_domain[\s\n]+/ && ( &set_routing_domain(), last TEST_SWITCH );

      #  Begin a test sequence.
      /^\s*sequence[\s\n]+/ && ( &begin_sequence, last TEST_SWITCH );

      #  Set up a user variable assignment.
      /^\s*setvar[\s\n]+/ && ( &set_variable, last TEST_SWITCH );

      #  Socket to use (idrp or ip)
      /^\s*socket[\s\n]+/ && ( &set_socket, last TEST_SWITCH );

      #  Name of the test.
      /^\s*test[\s\n]+/ && ( &begin_test, last TEST_SWITCH );

      #  Begin a test set.
      /^\s*test_set[\s\n]+/ && ( &begin_test_set, last TEST_SWITCH );

      #  Ignore comment lines.
      /^\s*#/ && last TEST_SWITCH;

      #  Ignore lines of whitespace only.
      /^\s*$/ && last TEST_SWITCH;

      #  Flag an error on other lines.
      &error_and_die( "UNKNOWN LINE\n\t$_" );
   }     # end TEST_SWITCH
}     # end while <TEST>

close( TEST );

#
#  Let the user know we are done.
#
print "==========================\n";
print "Test sets processed:   ",$#test_set_list + 1,"\n";
print "Sequences processed:   ",$num_sequences,"\n";
print "Total files generated: ",$num_cfg_files + $num_res_files,"\n";
print "\nProcessing complete.\n";

if ( $opt_I ) {
   close( NULL_OUT ); }

#
#====================  SUBROUTINES  ==============================
#


#
#  Subroutine:  READ_MACHINE_MAP
#  Read in information on machines.
#
sub read_machine_map
{
   local ( @static_routes, $route_list, $route );

   open( MACHINES, $machine_map_file )
      || die " *** ERROR - Cannot open machine map file: $!\n";

   print "Reading machine map file ... ";

   $i = 0;
   while ( <MACHINES> )  # assigns to Perl variable $_
   {
      LINE_SWITCH:
      {
         /^#/ && do              # comment line - ignore
         {
            last LINE_SWITCH;
         };

         /^\s*$/ && do              # blank line - ignore
         {
            last LINE_SWITCH;
         };

         # If we got here, it is a machine line.
         #
         ($machine[$i], $as_number[$i], $rdi[$i], $ip_addr_prefix[$i],
          $osi_addr_prefix[$i], $nlri_prefix[$i], $capacity[$i],
          $ip_address[$i], $osi_address[$i]) =
             split( /\s+/ );

         #  Make a copy of RDI/AS list for resetting after each sequence
         #   (rdi/as number can be overridden in a sequence)
         $save_rdi[$i] = $rdi[$i];
         $save_as[$i]  = $as_number[$i];

         if ( $osi_address[$i] =~ /\n$/ )
         {
            chop( $osi_address[$i] );   # Remove trailing newline.
         }

         $i++;
      }     # end LINE_SWITCH
   }     # end while

   if ( $#machine == 0 ) {
      print $#machine + 1," entry listed\n\n"; }
   else {
      print $#machine + 1," entries listed\n\n"; }

   close( MACHINES );
}

#
#  Subroutine:  CREATE_CONFIG_FILE
#  Create a gated configuration file for the given test name,
#  sequence, and subsequence.
#

sub create_config_file
{
   local ( $curr_mach_idx, $curr_test_set, @static_routes ) = @_;
   local ( $curr_seq, $config_file_name, $ribs, $key, $template, $fh_count );
   local ( @int_neighbors, @ext_neighbors );
   local ( $gateway, $rib, $route );
   local ( $sequence, $mach_idx );
   local ( $source_idx, $source );
   local ( %rib_list );

   #  Need to set for variable substitution (processing user setvars).
   $g001_mach_idx      = $curr_mach_idx;
   $g003_curr_test_set = $curr_test_set;
   $curr_seq           = $g002_curr_sequence;

   #  Generate the output configuration file name and open it.
   $config_file_name = "gated.conf.".$g021_curr_test_name."."
                       .$machine[$curr_mach_idx]."."
                       .$curr_seq.".".$curr_test_set;
   print "Generated configuration file name = $config_file_name\n";
   open( CONFIG_FILE, ">".$config_file_name )
      || die " *** ERROR - Cannot open configuration file: $! for output\n";

   #
   #  Process the template, filling in variables, just printing
   #  each line otherwise.
   #  Note: "*_" represents "$_" passed by reference.
   #
   $fh_count = -1;  # begin with main template
   while ( &get_next_line(*fh_count, *_) )
   {
      LINE_SWITCH:
      {
         #
         # The following are directives.  More than substitutions, these
         # specify a block of statements to generate.
         #
         /\$\$RIB_LOOP\b/ && do
         {
            #  Print the RIB template for every RIB in the list for this machine.
            &generate_rib_list( $g002_curr_sequence, $g001_mach_idx, *rib_list );
            while ( ($key,$template) = each %rib_list )
            {
               ($sequence, $mach_idx, $rib) = split( "$;", $key );
               &process_rib_template( $template, $rib );
            }
            last LINE_SWITCH;
         };

         /\$\$AREAS/ && do
         {
            $g006_curr_protocol = "ospf";
            &gen_area_statements( $curr_mach_idx );
            last LINE_SWITCH;
         };

         # Generate ISIS interface statements
         /\$\$ISIS_INTERFACES/ && do
         {
            $g006_curr_protocol = "isis";
            &gen_interfaces( $curr_mach_idx, $g007_curr_area );
            last LINE_SWITCH;
         };

         # Generate RIP interface statements
         /\$\$RIP_INTERFACES/ && do
         {
            $g006_curr_protocol = "rip";
            &gen_interfaces( $curr_mach_idx, $g007_curr_area );
            last LINE_SWITCH;
         };

         # Generate OSPF interface statements
         /\$\$OSPF_INTERFACES/ && do
         {
            $g006_curr_protocol = "ospf";
            &gen_interfaces( $curr_mach_idx, $g007_curr_area );
            last LINE_SWITCH;
         };

         /\$\$BGP_GROUPS/ && do
         {
            #  See if there is a template name included, and use it if
            #  there is no currently defined general group template name.
            #  If a group template has been specified for a particular
            #  protocol, that will override this value.
            $g006_curr_protocol = "bgp";
            &get_name( '\$\$BGP_GROUPS', $_, *template );
            if ( ($template ne "")
              && ($g170_template_list{group_template, general} eq "") )
            {
               $g170_template_list{group_template, general} = $template;
            }

            #  Generate group/peer statements for all appropriate peer types.
            &gen_peer_statements( "internal" );
            &gen_peer_statements( "external" );
            &gen_peer_statements( "igp" );
            &gen_peer_statements( "routing" );
            &gen_peer_statements( "test" );
            last LINE_SWITCH;
         };

         /\$\$IDRP_GROUPS/ && do
         {
            #  See if there is a template name included, and use it if
            #  there is no currently defined general group template name.
            #  If a group template has been specified for a particular
            #  protocol, that will override this value.
            $g006_curr_protocol = "idrp";
            &get_name( '\$\$IDRP_GROUPS', $_, *template );
            if ( ($template ne "")
              && ($g170_template_list{group_template, general} eq "") )
            {
               $g170_template_list{group_template, general} = $template;
            }

            #  Generate group/peer statements for all appropriate peer types.
            &gen_peer_statements( "internal" );
            &gen_peer_statements( "external" );
            last LINE_SWITCH;
         };

         #
         #  These are routes which the machine has declared static.
         /\$\$STATIC_ROUTES/ && do
         {
            foreach $route ( @g150_default_route_list, @static_routes )
            {
               print CONFIG_FILE "\t$route";
            }
            last LINE_SWITCH;
         };

         /\$\$IMPORT_STATEMENTS/ && do
         {
#           print "Processing import statements...\n";
            &process_import_list( $curr_mach_idx );
            last LINE_SWITCH;
         };

         /\$\$EXPORT_STATEMENTS/ && do
         {
#           print "Processing export statements...\n";
            &process_export_list( $curr_mach_idx );
            last LINE_SWITCH;
         };

         # Process general, non-directive predefined variables.
         &process_predef_vars;  # will use $_ as line to check

         # Process user-defined variables.
         &process_user_vars;    # will use $_ as line to check

         print CONFIG_FILE $_;
      }

   }  # end while

   close ( CONFIG_FILE );
}

#
#  Subroutine:  GET_NEXT_LINE
#   Read the next line from the current template.
#   Handles $$INCLUDE directives through recursive calls.
#
sub get_next_line
{
   local ( *fh_count, *next_line ) = @_;
   local ( $exp_filehandle, $return_code );
   local ( $dummy, $inc_file_name );

   # Special case: opening the first file (main template).
   if ( $fh_count == -1 )
   {
      &find_file( *main_template_file )
         || die " *** ERROR - Cannot find main template file \"$main_template_file\" : $!\n";

      $exp_filehandle = "INCL_TEMPLATE_".++$fh_count;
      open( $exp_filehandle, $main_template_file )
         || die " *** ERROR - Cannot open template file \"$main_template_file\" : $!\n";
   }
   else
   {
      # Set filehandle to current value.
      $exp_filehandle = "INCL_TEMPLATE_".$fh_count;
   }

   # Is this the last line of this file?
   if ( eof($exp_filehandle) )
   {
      # If this is the last or only file on the stack, we are all done.
      close( $exp_filehandle );
      return(0)  if ( --$fh_count == -1 );

      # Not done -- go back to the previous file and keep reading.
      $return_code = &get_next_line( *fh_count, *next_line );
   }
   else
   {
      $next_line = <$exp_filehandle>;
      $return_code = 1;
   }

   # Is this line an INCLUDE directive?
   if ( $next_line =~ /^\s*\$\$INCLUDE\s+/ )
   {
      chop ( $next_line );
      ( $dummy, $_ ) = split( /^\s*\$\$INCLUDE\s+/, $next_line );

      # The file name to include could be a user variable.
      &process_user_vars;   # uses $_ as line to process
      $inc_file_name = $_;

      # Find the full path of the file name.
      &find_file( *inc_file_name )
         || die " *** ERROR - Cannot find include file \"$inc_file_name\": $!\n";

      $exp_filehandle = "INCL_TEMPLATE_".++$fh_count;
      open( $exp_filehandle, $inc_file_name )
         || die " *** ERROR - Cannot open include file \"$inc_file_name\": $!\n";
      $return_code = &get_next_line( *fh_count, *next_line );
   }

   # If we got this far, we have a line to return.
   return($return_code);
}

#
#  Subroutine:  GEN_AREA_STATEMENTS
#
sub gen_area_statements
{
   local( $host_idx ) = @_;
   local( $protocol, $machine_name, $area, $interface, $template );
   local( $entry, $host, $peer, $peer_idx, $area_idx );
   local( @area_list ) = ();

   $host = $machine[$host_idx];
   foreach $entry ( @g100_data_list )
   {
      ( $protocol, $machine_name, $area, $interface, $template ) =
         split( "$;", $entry );

      # Generate a unique list of areas for this machine in current sequence.
      if ( ($protocol eq "ospf") && ($host eq $machine_name) ) {
         &push_unique( $area, *area_list ); }
   }

   #  Generate code from area template for each area.
   foreach $area ( @area_list )
   {
      $area_idx = &find_machine( $area );
      $template = $g112_area_template_list{$area};
      open( AREA_TEMPLATE, $template )
         || die " *** ERROR - Cannot open area template file \"$template\" : $!\n";

      while( <AREA_TEMPLATE> )
      {
         # Generate the interface list if this line is the directive.
         if ( /\$\$OSPF_INTERFACES/ )
         {
            &gen_interfaces( $host_idx, $area );
         }
         else
         {
            # Process area-specific, non-directive predefined variables.
            s/\$\$AREA_NAME/$ip_address[$area_idx]/;

            # Process general, non-directive predefined variables.
            &process_predef_vars;  # will use $_ as line to check

            # Process user-defined variables.
            &process_user_vars;    # will use $_ as line to check

            print CONFIG_FILE $_;
         }
      }
      close( AREA_TEMPLATE );
   }     # foreach
}


#
#  Subroutine:  GEN_INTERFACES
#
sub gen_interfaces
{
   local( $host_idx, $curr_area ) = @_;
   local( $protocol, $machine_name, $area, $interface, $template );
   local( $entry, $host, $intf_idx, $intf_address );

   $host = $machine[$host_idx];
   foreach $entry ( @g100_data_list )
   {
      ( $protocol, $machine_name, $area, $interface, $template ) =
         split( "$;", $entry );

      # Generate interfaces for entries which match the current protocol and
      # machine name.  If the protocol is OSPF, the areas must also match.
      if ( ($protocol eq $g006_curr_protocol)
        && ($machine_name eq $host) )
      {
         next if ( ($protocol eq "ospf") && ($area ne $curr_area) );

         open( INTF_TEMPLATE, $template )
            || die " *** ERROR - Cannot open interface template file \"$template\" : $!\n";

         # The "all" keyword is a special case.  Let it through untouched.
         if ( $interface eq "all" ) {
            $intf_address = "all"; }
         else {
            $intf_address = $ip_address[ &find_machine($interface) ]; }

         # Set this for user variable translation (process_user_vars).
         $g013_curr_peer_intf = $interface;

         while( <INTF_TEMPLATE> )
         {
            # Process interface-specific, non-directive predefined variables.
            s/\$\$INTERFACE_IP/$intf_address/;

            # Process general, non-directive predefined variables.
            &process_predef_vars;  # will use $_ as line to check

            # Process user-defined variables.
            &process_user_vars;    # will use $_ as line to check

            print CONFIG_FILE $_;
         }
         close( INTF_TEMPLATE );
      }     # if
   }     # foreach
}

#
#  Subroutine:  GEN_PEER_STATEMENTS
#   Generates group and peer statements for IDRP and BGP.
#
sub gen_peer_statements
{
   local( $peer_type_to_gen ) = @_;
   local( $protocol, $machine_name, $peer_type, $template );
   local( $entry, $host, $peer, $peer_idx, $as, $dummy );
   local( @bgp_ext_list, @peer_list ) = ();

   $host_idx = $g001_mach_idx;
   $host = $machine[$host_idx];

   #  BGP external peers represent a special case; each different AS needs
   #  its own external group statement.
   if ( ($g006_curr_protocol eq "bgp") && $peer_type_to_gen eq "external" )
   {
      # Extract a list of the external AS numbers in the data list.
      foreach $entry ( @g100_data_list )
      {
         ( $protocol, $machine_name, $peer_type, $peer, $template ) =
            split( "$;", $entry );
         if ( ($peer_type eq "external") && ($protocol eq "bgp")
           && ($machine_name eq $host) )
         {
            $peer_idx = &find_machine( $peer );
            &push_unique( $as_number[$peer_idx], *as_list );
            push( @bgp_ext_list, $entry );
         }
      }

      # Now for each unique AS number in the list, extract a list of external
      # peers with this AS number and process it as a group.
      foreach $as ( @as_list )
      {
         @peer_list = ();
         foreach $entry ( @bgp_ext_list )
         {
            ( $protocol, $machine_name, $peer_type, $peer, $template ) =
               split( "$;", $entry );
            $peer_idx = &find_machine( $peer );
            if ( $as_number[$peer_idx] eq $as ) {
               push( @peer_list, $entry ); }
         }

         # Generate a group with this list of peers.
         if ( $#peer_list > -1 ) {
            &gen_group( $host_idx, $peer_type_to_gen, $as, @peer_list ); }
      }
   }   # if external BGP

   #
   #  IDRP peer groups and non-external BGP groups.
   else
   {
      # Build a sublist of peer entries which belong to the current machine,
      # protocol, and peer type.
      foreach $entry ( @g100_data_list )
      {
         ( $protocol, $machine_name, $peer_type, $peer, $template ) =
            split( "$;", $entry );

         if (  ($protocol eq $g006_curr_protocol)
            && ($peer_type eq $peer_type_to_gen)
            && ($machine_name eq $host) ) {
            push( @peer_list, $entry ); }
      }

      # If there are any entries in the sublist, open the group template
      # and process it and the peers.
      if ( $#peer_list > -1 ) {
         &gen_group( $host_idx, $peer_type_to_gen,
                     $as_number[$host_idx], @peer_list ); }
   }   # else
}

#
#  Subroutine:  GEN_GROUP
#   Generates a group statement for IDRP and BGP.
#
sub gen_group
{
   local( $host_idx, $peer_type, $peer_as, @peer_list ) = @_;
   local( $entry, $template, $loop_template );

   if ( $template eq "" ) {
      $template = $g170_template_list{group_template, $g006_curr_protocol}; }
   if ( $template eq "" ) {
      $template = $g170_template_list{group_template, general}; }
   if ( $template eq "" ) {
      die "***** ERROR -- No group template defined\n"; }

   &find_file( *template )
     || die " *** ERROR - Cannot find group template file \"$template\" : $!\n";
   open( GROUP_TEMPLATE, $template )
     || die " *** ERROR - Cannot open group template file \"$template\" : $!\n";

   while( <GROUP_TEMPLATE> )
   {
      # Process group-specific directives.
      if ( /\$\$PEER_LOOP\b/ )
      {
         # See if there is a template name included.
         &get_name( '\$\$PEER_LOOP', $_, *loop_template );

         # Process each peer in the list.
         foreach $entry ( @peer_list ) {
            &process_peer( $entry, $host_idx, $loop_template ); }

         #  Skip over the user variable processing and do *not* print the
         #  $$PEER_LOOP line in the file.
         next;
      }  # $$PEER_LOOP code

      # Process group-specific, non-directive predefined variables.
      s/\$\$GROUP_TYPE/$peer_type/;
      s/\$\$PEER_AS/$peer_as/;
      s/\$\$HOST_AS/$as_number[$host_idx]/;
      s/\$\$HOST_RDI/$rdi[$host_idx]/;
      s/\$\$HOST_IP/$ip_address[$host_idx]/;
      s/\$\$HOST_NET/$osi_address[$host_idx]/;

      # Process general, non-directive predefined variables.
      &process_predef_vars;  # will use $_ as line to check

      # Process user-defined variables.
      &process_user_vars;    # will use $_ as line to check

      print CONFIG_FILE $_;
   }
   close( GROUP_TEMPLATE );
}

#
#  Subroutine:  PROCESS_PEER
#   Generates a peer statement for IDRP and BGP.
#
sub process_peer
{
   local( $entry, $host_idx, $loop_template ) = @_;
   local( $protocol, $machine_name, $peer_type, $peer, $template );
   local( $peer_idx, $ribs );
   local( $peer_idx );

   ( $proto, $machine_name, $peer_type, $peer, $template ) =
      split( "$;", $entry );

   $peer_idx = &find_machine( $peer );
   $ribs = $peer_ribs_list{$g002_curr_sequence, $host_idx, $peer};
   if ( $ribs eq "" ) { $ribs = "all"; }

   # Set this for user variable translation (process_user_vars).
   $g013_curr_peer_intf = $peer;

   # The "template" field of an entry overrides the template
   # name specified in the $$PEER_LOOP variable.
   if ( $template eq "" ) {
      $template = $loop_template; }
   if ( $template eq "" ) {
      die "***** ERROR -- No peer sub-template listed\n"; }

   open( PEER_TEMPLATE, $template )
      || die " *** ERROR - Cannot open peer sub-template file \"$template\" : $!\n";

   while( <PEER_TEMPLATE> )
   {
      # Process peer-specific, non-directive predefined variables.
      s/\$\$PEER_NAME/$peer/;
      s/\$\$PEER_NET/$osi_address[$peer_idx]/;
      s/\$\$PEER_IP/$ip_address[$peer_idx]/;
      s/\$\$PEER_AS/$as_number[$peer_idx]/;
      s/\$\$PEER_RDI/$rdi[$peer_idx]/;
      s/\$\$GROUP_TYPE/$peer_type/;
      s/\$\$HOST_AS/$as_number[$host_idx]/;
      s/\$\$HOST_RDI/$rdi[$host_idx]/;
      s/\$\$HOST_IP/$ip_address[$host_idx]/;
      s/\$\$HOST_NET/$osi_address[$host_idx]/;
      s/\$\$RIBS-SUPPORTED/$ribs/;

      # Process general, non-directive predefined variables.
      &process_predef_vars;  # will use $_ as line to check

      # Process user-defined variables.
      &process_user_vars;    # will use $_ as line to check

      print CONFIG_FILE $_;
   }
   close( PEER_TEMPLATE );
}

#
#  Subroutine:  READ_DEFAULT_ROUTES
#   Read the route set file and read in the set of default routes listed.
#
sub read_default_routes
{
   local ( @route_list, $route_set_found, $end_found, $count );
   local ( $route_set_id, $def_route_set_id );

   $def_route_set_id = $g140_route_set_id_list{$g003_curr_test_set, "default"};

   # Every test set SHOULD have a default route, but there may be cases where
   # one isn't wanted, so print a warning.
   if ( $def_route_set_id eq "" )
   {
      if ( $print_warnings eq ON ) {
         print "WARNING:  No default route set defined for test set $g003_curr_test_set\n"; }
      return;
   }

   # Open the route set file.
   open( ROUTE_SET_FILE, $route_set_file )
      || die " *** ERROR - Cannot open route set file \"$route_set_file\": $!\n";

   # Find the appropriate default route set in the route set file.
   $route_set_found = $FALSE;
   while ( <ROUTE_SET_FILE> )
   {
      if ( /^default-routes/ )
      {
         &get_name( 'default-routes', $_, *route_set_id )
             || &error_and_die( "No ID specified for \"default-routes\" in route set file" );
         if ( $route_set_id eq $def_route_set_id )
         {
            $route_set_found = $TRUE;
            last;
         }
      }
   }

   if ( $route_set_found == $FALSE ) {
      &error_and_die( "No default route set \"$def_route_set_id\" defined in route set file" ); }

   # Read all default routes.
   $count = 0;
   $end_found = $FALSE;
   LINE: while ( <ROUTE_SET_FILE> )
   {
      next LINE if ( (/^#/) || (/^\s*$/) );  # comment or blank line - ignore

      /^end/ && do                           # end of default route set
      {
         $end_found = $TRUE;
         last LINE;
      };

      # This must be a default route.  Substitute for any predefined
      # or user-defined variables, then add to list.
      &process_predef_vars;  # will use $_ as line to check
      &process_user_vars;    # will use $_ as line to check
      $route_list[$count++] = $_;
   }
   close ( ROUTE_SET_FILE );

   if ( $end_found == $FALSE )
   {
      if ( $print_warnings eq ON ) {
         print "WARNING:  no \"end\" statement found for default route set\n"; }
   }

   # Return the list.
   return ( @route_list );
}

#
#  Subroutine:  READ_STATIC_ROUTES
#   Read the route set file and get the appropriate set of static routes
#   based on machine name and route set name.
#
sub read_static_routes
{
   local ( $test_set, $mach_idx ) = @_;
   local ( @route_list, $route_set_id, $route_set_name, $machine_name );
   local ( $route_set_found, $end_found, $count );

   @route_list = ();

   # Open the route set file.
   open( ROUTE_SET_FILE, $route_set_file )
      || die " *** ERROR - Cannot open route set file \"$route_set_file\": $!\n";

   # Find out which route set to use.
   $route_set_id = $g140_route_set_id_list{$test_set, $mach_idx};

   # If no machine specified (e.g. "route_set 1", use "all" setting.
   # This means use the specified route set for every machine.
   if ( $route_set_id eq "" ) {
      $route_set_id = $g140_route_set_id_list{$test_set, "all"}; }

   # If using "null" route set (no static routes), just return null list.
   if ( $route_set_id eq "null" ) {
      return( @route_list ); }

   # If no machine specified and no "all" setting, assume "null", but warning.
   if ( $route_set_id eq "" )
   {
      if ( $print_warnings eq ON ) {
         print "WARNING:  no route set specified for machine $machine[$mach_idx] in test set $test_set\n"; }
      return( @route_list );
   }

   # Find the route set in the route set file.
   $route_set_found = $FALSE;
   while ( <ROUTE_SET_FILE> )
   {
      if ( /^$machine[$mach_idx]/ )
      {
         ( $machine_name, $route_set_name ) = split( /\s+/ );
         if ( $route_set_name eq $route_set_id )
         {
            $route_set_found = $TRUE;
            last;
         }
      }
   }

   if ( $route_set_found == $FALSE ) {
      &error_and_die( "Machine $machine[$mach_idx] does not have route set \"$route_set_id\" defined in route set file" ); }

   # Read all routes in that route set.
   $count = 0;
   $end_found = $FALSE;
   LINE: while ( <ROUTE_SET_FILE> )
   {
      next LINE if ( (/^#/) || (/^\s*$/) );  # comment or blank line - ignore

      /^end/ && do                           # end of route set
      {
         $end_found = $TRUE;
         last LINE;
      };

      # This must be a static route in current route set.
      # Substitute for any user-defined variables, then add to list.
      &process_predef_vars;  # will use $_ as line to check
      &process_user_vars;    # will use $_ as line to check
      $route_list[$count++] = $_;
   }
   close ( ROUTE_SET_FILE );

   if ( $end_found == $FALSE )
   {
      if ( $print_warnings eq ON ) {
         print "WARNING:  no \"end\" statement found for route set $route_set_name\n"; }
   }

   # Return the list.
   return ( @route_list );
}

#
#  Subroutine:  PROCESS_RIB_TEMPLATE
#
sub process_rib_template
{
   local ( $rib_template, $rib ) = @_;
   local ( $token, $value );

#  print "Current RIB = $rib\n";

   open( RIB_TEMPLATE, $rib_template )
      || die " *** ERROR - Cannot open rib template file \"$rib_template\": $!\n";

   while ( <RIB_TEMPLATE> )
   {
      #
      # If a match is made with the following set of variables,
      # make the appropriate substitution.
      #
      if ( s/\$RIB\b/$rib/ )
      {
#        print "*** Substitution: New line = $_";
      }

      # Process general, non-directive predefined variables.
      &process_predef_vars;  # will use $_ as line to check

      # Process user-defined variables.
      &process_user_vars;    # will use $_ as line to check

      print CONFIG_FILE $_;
   }
   close ( RIB_TEMPLATE );
}

#
#  Subroutine:  GENERATE_RIB_LIST
#  Generate a list of RIBs for a given sequence/machine pair.
#
sub generate_rib_list
{
    local ( $curr_sequence, $curr_mach_idx, *generated_rib_list ) = @_;
    local ( $key, $template, $sequence, $mach_idx, $rib );
    local ( %remaining_rib_list );
 
    %generated_rib_list = ();
    %remaining_rib_list = ();

    #
    #  Look through the entire set of defined RIBs.  Start with those
    #  which match globally, override if a local definition redefines it.
    #
    while ( ($key,$template) = each %entire_ribs_list )
    {
       #  Split key into its components (assoc. array separator is $;).
       ($sequence, $mach_idx, $rib) = split( "$;", $key );
 
       if ( $sequence eq "$ALL" )
       {
          if ( $mach_idx ne "$ALL" ) {
              &error_and_die( "Unknown RIB configuration" ); }
          $generated_rib_list{$curr_sequence, $curr_mach_idx, $rib} = $template;
#         print "*** Global:  Handling RIB $sequence, $mach_idx, $rib\n";
       }
       else
       {
          $remaining_rib_list{$sequence, $mach_idx, $rib} = $template;
       }
    }
 
    #  Now process sequence scope RIBs.
    while ( ($key,$template) = each %remaining_rib_list )
    {
       #  Split key into its components (assoc. array separator is $;).
       ($sequence, $mach_idx, $rib) = split( "$;", $key );
 
       if ( ($sequence eq $curr_sequence) && ($mach_idx eq "$ALL") )
       {
          $generated_rib_list{$curr_sequence, $curr_mach_idx, $rib} = $template;
#         print "*** Sequence:  Handling RIB $sequence, $mach_idx, $rib\n";
       }
    }
 
    #  Now process machine scope RIBs.
    while ( ($key,$template) = each %remaining_rib_list )
    {
       #  Split key into its components (assoc. array separator is $;).
       ($sequence, $mach_idx, $rib) = split( "$;", $key );
 
       if ( ($sequence eq $curr_sequence) && ($mach_idx eq $curr_mach_idx) )
       {
          $generated_rib_list{$curr_sequence, $curr_mach_idx, $rib} = $template;
#         print "*** Machine:  Handling RIB $sequence, $mach_idx, $rib\n";
       }
    }

#   while ( ($key,$template) = each %generated_rib_list )
#   {
#      ($sequence, $mach_idx, $rib) = split( "$;", $key );
#      print "--- Generated RIB list entry = $sequence, $mach_idx, $rib\n";
#   }

}

#
#  Subroutine:  PROCESS_PREDEF_VARS
#  Substitute for predefined ("$$") variables which are not directives.
#
sub process_predef_vars
{
   #  Information usually found in the main template's header.
   s/\$\$DATE/$date/;
   s/\$\$HOST_NAME/$machine[$g001_mach_idx]/;
   s/\$\$TEST_NAME/$g021_curr_test_name/;
   s/\$\$SEQUENCE/$g002_curr_sequence/;
   s/\$\$TEST_SET/$g003_curr_test_set/;
   s/\$\$PROGRAM_NAME/$PROGRAM_NAME/;
   s/\$\$PROGRAM_VERSION/$VERSION/;

   #  Information usually found in a protocol-specific template.
   s/\$\$TRACEOPTIONS/\"$log_file_dir\/$log_file_name\"/;
   s/\$\$RDI_NUMBER/$rdi[$g001_mach_idx]/;
   s/\$\$AS_NUMBER/$as_number[$g001_mach_idx]/;
   s/\$\$NET/$osi_address[$g001_mach_idx]/;
   s/\$\$CAPACITY/$capacity[$g001_mach_idx]/;
   s/\$\$PROTO-SOCK/$g031_proto_sock/;
}

#
#  Subroutine:  PROCESS_USER_VARS
#
sub process_user_vars
{
   local ( $token, $value );

#  print "*** process_user_vars:  Line = \"$_\"\n";

   if ( /\$/ )      # compares against $_
   {
      foreach $token ( split(/[\s;]+/) )
      {
#        print "*** process_user_vars:  Token = \"$token\"\n";
         if ( $token =~ /^\$/ )
         {
            $value = &find_variable( $token );
            $token = "\\".$token;
            s/$token/$value/;   # replaces token in $_
#           print "*** Substitution: New line = $_";
         }
      }
   }
}

#
#  Subroutine:  PROCESS_IMPORT_LIST
#
sub process_import_list
{
   local ( $mach_idx ) = @_;
   local ( $order, $test_set, $protocol, $type, $index );
   local ( $key, $template );
   local ( @peer_list, $peer );

   #  Sorting the keys insures the array entries are returned in the
   #  order in which they were added (because the first part of the
   #  key is a large incrementing number).
   foreach $key (sort(keys %g120_import_stmt_list))
   {
      $template = $g120_import_stmt_list{$key};
      ($order, $test_set, $protocol, $type, $index) = split( "$;", $key );
#     print "Import test set array{ $order, $test_set, $protocol, $type, $index } = $template\n";

      if ( $test_set eq $g003_curr_test_set )
      {
         TYPE_SWITCH:
         {
            ( $type eq "LOOP" ) && do
            {
               # Get a list of the peers of the current machine
               # which match the necessary protocol.
               @peer_list = &gen_rdi_list( $mach_idx, $protocol );

               # Process each peer in the list.
               foreach $peer ( @peer_list ) {
                  &process_import_template( $peer, $template ); }
               last TYPE_SWITCH;
            };

            ( $type eq "SINGLE" ) && do
            {
               &process_import_template( "", $template );
               last TYPE_SWITCH;
            };

            &error_and_die( "Unknown import type" );
         }   # end TYPE_SWITCH
      }   # end if
   }   # end foreach
}

#
#  Subroutine:  PROCESS_IMPORT_TEMPLATE
#
sub process_import_template
{
   local ( $peer, $imp_template ) = @_;
   local ( $token, $value );

   open( IMP_TEMPLATE, $imp_template )
      || die " *** ERROR - Cannot open import template file \"$imp_template\": $!\n";

   while ( <IMP_TEMPLATE> )
   {
      #
      # Process import-specific, non-directive predefined variables.
      s/\$\$PEER_AS\b/$as_number[$peer]/;
      s/\$\$PEER_RDI\b/$rdi[$peer]/;

      # Process general, non-directive predefined variables.
      &process_predef_vars;  # will use $_ as line to check

      # Process user-defined variables.
      &process_user_vars;    # will use $_ as line to check

      print CONFIG_FILE $_;
   }
   close ( IMP_TEMPLATE );
}

#
#  Subroutine:  PROCESS_EXPORT_LIST
#
sub process_export_list
{
   local ( $mach_idx ) = @_;
   local ( $order, $test_set, $type, $index );
   local ( $key, $template );
   local ( @peer_list, $peer );

   #  Sorting the keys insures the array entries are returned in the
   #  order in which they were added (because the first part of the
   #  key is a large incrementing number).
   foreach $key (sort(keys %g130_export_stmt_list))
   {
      $template = $g130_export_stmt_list{$key};
      ($order, $test_set, $protocol, $type, $index) = split( "$;", $key );
#     print "Export test set array{ $order, $test_set, $protocol, $type, $index } = $template\n";

      if ( $test_set eq $g003_curr_test_set )
      {
         TYPE_SWITCH:
         {
            ( $type eq "LOOP" ) && do
            {
               # Get a list of the peers of the current machine
               # which match the necessary protocol.
               @peer_list = &gen_rdi_list( $mach_idx, $protocol );

               # Process each peer in the list.
               foreach $peer ( @peer_list ) {
                  &process_export_template( $peer, $protocol, $template ); }
               last TYPE_SWITCH;
            };

            ( $type eq "SINGLE" ) && do
            {
               &process_export_template( "", $protocol, $template );
               last TYPE_SWITCH;
            };

            &error_and_die( "Unknown export type" );
         }   # end TYPE_SWITCH
      }   # end if
   }   # end foreach
}

#
#  Subroutine:  PROCESS_EXPORT_TEMPLATE
#
sub process_export_template
{
   local ( $peer, $protocol, $exp_template ) = @_;
   local ( $sub_template, @peer_list, $peer2, $count );
   local ( $token, $value );
   local ( $exp_filehandle );

   # Find the full path of the export template.
   &find_file( *exp_template )
      || die " *** ERROR - Cannot find template file \"$exp_template\": $!\n";

   # Calculate a new filehandle name each time so we don't step on it
   #  when doing recursive calls to this routine.
   $exp_filehandle = "EXP_TEMPLATE_".$gl_exp_count++;
   open( $exp_filehandle, $exp_template )
      || die " *** ERROR - Cannot open export template file \"$exp_template\": $!\n";

   $count = 0;
   while ( <$exp_filehandle> )
   {
      #
      # If a match is made with the following set of variables,
      # make the appropriate substitution.
      #
      s/\$\$PEER_AS\b/$as_number[$peer]/;
      s/\$\$PEER_RDI\b/$rdi[$peer]/;
      s/\$\$HOST_AS/$as_number[$g001_mach_idx]/;
      s/\$\$HOST_RDI/$rdi[$g001_mach_idx]/;

      if ( /\$\$LOOP\b/ )
      {
         &get_name( '\$\$LOOP', $_, *sub_template )
             || die "***** ERROR -- No sub-template listed\n";

         # Get a list of the peers of the current machine
         # which match the necessary protocol.
         @peer_list = &gen_rdi_list( $g001_mach_idx, $protocol );

         # Process each peer in the list.
         foreach $peer2 ( @peer_list )
         {
            #
            # Process the "routes imported from" nodes.
            # If loop code is "off", and this is the host machine, skip it.
            if ( ($rdi[$g001_mach_idx] ne $rdi[$peer2])
              || ($g202_opt_gen_loop_code eq "on") )
            {
#              print "Peer 2 RDI = $rdi[$peer2]\n";
               &process_export_template( $peer2, $protocol, $sub_template );
            }
         }

         #  Skip over the user variable processing and do *not* print the
         #  $$LOOP line in the file.
         next;
      }

      # Process general, non-directive predefined variables.
      &process_predef_vars;  # will use $_ as line to check

      # Process user-defined variables.
      &process_user_vars;    # will use $_ as line to check

      print CONFIG_FILE $_;
   }
   close ( $exp_filehandle );
}

#
#  Subroutine:  FIND_VARIABLE
#  The config file allows for user setting of pseudo-variables.
#  Look up the value of one of these variables in the setvar tables.
#
sub find_variable
{
   local ( $token ) = @_;
   local ( $value );

   # See if user has set it globally.  Override predefined.
   if ( $global_var_list{$token} ne "" ) {
      $value = $global_var_list{$token}; }

   # See if user has set it inside the current sequence.  Override global.
   if ( $sequence_var_list{$g002_curr_sequence, $token} ne "" ) {
      $value = $sequence_var_list{$g002_curr_sequence, $token}; }

   # See if user has set it inside the current machine.  Override sequence.
   if ( $machine_var_list{$g002_curr_sequence, $g001_mach_idx, $token} ne "" ) {
      $value = $machine_var_list{$g002_curr_sequence, $g001_mach_idx, $token} };

   # See if user has set it for a peer or interface.  Override machine.
   if ( $peer_var_list{$g002_curr_sequence, $g001_mach_idx, $g013_curr_peer_intf, $token} ne "" ) {
      $value = $peer_var_list{$g002_curr_sequence, $g001_mach_idx, $g013_curr_peer_intf, $token} };

   # See if user has set it inside the current test set.  Override machine.
   if ( $test_set_var_list{$test_set, $token} ne "" ) {
      $value = $test_set_var_list{$test_set, $token}; }

   if ( $value eq "" )
   {
      if ( $print_warnings eq ON ) {
      print "WARNING:  \"$token\" has no value\n"; }
   }

#  print "find_variable: token = $token\n";
#  print "find_variable: value = $value\n";

   return( $value );
}

#
#  Subroutine:  GEN_RDI_LIST
#  Generate a list of unique RDI of host machine and peers.  Return a list
#  of machines with these RDI numbers.
#
sub gen_rdi_list
{
   local ( $mach_idx, $protocol_wanted ) = @_;
   local ( $entry, $peer_idx, $duplicate );
   local ( @tmp_mach_list, $tmp_mach );
   local ( @mach_list, $new_mach );
   local ( $protocol, $machine_name, $peer_type, $peer, $template );
   @tmp_mach_list = ();
   @mach_list = ();

   #  Create a list of machines representing all RDI in the peer list.
   foreach $entry ( @g100_data_list )
   {
      ( $protocol, $machine_name, $peer_type, $peer, $template ) =
         split( "$;", $entry );

      if ( ($protocol eq $protocol_wanted)
           && ($machine_name eq $machine[$mach_idx]) )
      {
         $peer_idx = &find_machine( $peer );
         $tmp_mach_list[$#tmp_mach_list + 1] = $peer_idx;
      }
   }

   #  Add the host machine (to get host machine RDI).
   $tmp_mach_list[$#tmp_mach_list + 1] = $mach_idx;

   #  Now eliminate duplicates.
   foreach $tmp_mach ( @tmp_mach_list )
   {
      $duplicate = 0;
      foreach $new_mach ( @mach_list )
      {
         if ( $rdi[$tmp_mach] eq $rdi[$new_mach] ) {
            $duplicate = 1; }
      }
      if ( ! $duplicate ) {
         $mach_list[$#mach_list + 1] = $tmp_mach; }
   }

#  for ( $i = 0; $i <= $#mach_list; $i++ )
#  {
#     print "gen_rdi_list: mach_list[$i] = $mach_list[$i]\n";
#  }

   return ( @mach_list );
}

#
#  Subroutine:  CREATE_RESULT_FILE
#  Create a file listing the routes expected to be in the
#  routing table for a particular test, sequence, and test set.
#
sub create_result_file
{
   local ( $curr_mach_idx, $curr_test_set, @static_routes ) = @_;
   local ( $curr_seq, $i, $mach, $rib, $route );
   local ( $key, $sequence, $mach_idx );
   local ( $result_file_name );
   local ( @int_neighbors, @ext_neighbors, @mach_list, %rib_list );
   local ( @peer_ribs_supported, @supported_list );
   local ( $protocol, $machine_name, $peer_type, $peer, $template );
   local ( @route_list, @peer_static_routes );

   $curr_seq = $g002_curr_sequence;

   $result_file_name = "expected.".$g021_curr_test_name."."
                       .$machine[$curr_mach_idx]."."
                       .$curr_seq.".".$curr_test_set;
   print "Expected results file name = $result_file_name\n";

   #
   #  Open the output results file.
   #
   open( RESULT_FILE, ">".$result_file_name )
      || die " *** ERROR - Cannot open expected results file: $! for output\n";

   #
   #  Push the default routes onto the expected routes list.
   #
   foreach $route ( @g150_default_route_list ) {
      &push_unique( $route, *route_list ); }

   #
   #  Add the IP and ISO routes of the currrent machine and its neighbors.
   #
   @mach_list  = ( $machine[$curr_mach_idx] );
   foreach $entry ( @g100_data_list )
   {
      ( $protocol, $machine_name, $peer_type, $peer, $template ) =
         split( "$;", $entry );

      # Don't process entries with peer name "all".  This is an OSPF/RIP
      # interface entry and has no meaning here.
      if ( ($machine_name eq $machine[$curr_mach_idx]) && ($peer ne "all") )
      {
         &push_unique( $peer, *mach_list );
      }
   }

   # Read static routes for each peer.  Add each one to the static route list,
   # eliminating duplicates.
   foreach $route ( @static_routes ) {
      &push_unique( $route, *route_list ); }
   foreach $peer ( @mach_list )
   {
      $mach_idx = &find_machine( $peer );
      @peer_static_routes = &read_static_routes( $curr_test_set, $mach_idx );
      foreach $route ( @peer_static_routes ) {
         &push_unique( $route, *route_list ); }
   }

   #  Create a unique list of all "ribs-supported" by this machine's peers.
   #  If any say "ribs-supported all", use all RIBs defined for this machine.
   foreach $peer ( @mach_list )
   {
      #  Always use the default RIB.
      @peer_ribs_supported = ( $default_rib ) ;

      #  Add specific RIBs for this peer.
      push( @peer_ribs_supported, split( /\s+/,
           $peer_ribs_list{$curr_seq, $curr_mach_idx, $peer}) );

      @supported_list = ();
      foreach $rib ( @peer_ribs_supported )
      {
         if ( $rib eq "all" )
         {
            # Generate a list of all RIBS supported by this machine.
            &generate_rib_list( $curr_seq, $curr_mach_idx, *rib_list );
            while ( ($key,$template) = each %rib_list )
            {
               ($sequence, $mach_idx, $rib) = split( "$;", $key );
               push( @supported_list, $rib );
            }
         }
         else
         {
            push( @supported_list, $rib );
         }
      }
   }

   #  Put each supported RIB into expected results file.
   foreach $rib ( @supported_list )
   {
      # Print static routes for each RIB.
      foreach $route ( @route_list )
      {
         print RESULT_FILE "$rib\t$route";
      }
   }     # while loop - each rib

   close ( RESULT_FILE );
}

#
#  Subroutine:  ADD_GENERATE_ENTRY
#    Add restriction to sequence/test set generation.
#    Format is: { <sequence ID | *> , <test set ID | *> } | "all"
#      "*" is the wildcard - matches anything.
#      "all" is equivalent to "*,*"
#
#    This list uses the most inclusive set:  thus if there are entries
#      1,*
#      2,2
#      all
#    every combination will be generated.
#
sub add_generate_entry
{
   local ( $entry );

   &get_name( 'generate', $_, *entry )
       || &error_and_die( "No value in generate statement" );

   push( @g160_generation_list, $entry );
}

#
#  Subroutine:  BEGIN_TEST
#
sub begin_test
{
   @test_set_list = ();

   &get_name( 'test', $_, *g021_curr_test_name )
       || &error_and_die( "Test has no name" );
   print "Now beginning test $g021_curr_test_name ...\n";
}

#
#  Subroutine:  SET_AREA
#
#    Sets the currently active OSPF area inside a sequence.
#
sub set_area
{
   local ( $area_idx, $template );

   if ( $g006_curr_protocol ne "ospf" ) {
      &error_and_die( "Area keyword valid only for OSPF protocol" ); }

   &get_name( 'area', $_, *g007_curr_area )
      || &error_and_die( "No area ID specified" );

   # Get a template name if one is specified.  Otherwise set to current.
   ( $g007_curr_area, $template ) = split( "=", $g007_curr_area );
   if ( $template eq "" ) {
      $template = $g170_template_list{area_template, $g006_curr_protocol}; }
   if ( $template eq "" ) {
      $template = $g170_template_list{area_template, general}; }
   if ( $template eq "" ) {
      die "***** ERROR -- No area template defined\n"; }

   # Find the full path of the file name.
   &find_file( *template )
      || die " *** ERROR - Cannot find template file \"$template\": $!\n";

   # Make sure area name has a definition in the table.
   $area_idx = &find_machine( $g007_curr_area );
   if ( $area_idx == -1 ) {
      &error_and_die( "Area \"$g007_curr_area\" not found in table" ); }

   # Assign the template to the area.  (You can have a different area
   # template for each area, since each area has a unique name.)
   $g112_area_template_list{$g007_curr_area} = $template;
}

#
#  Subroutine:  SET_TEMPLATE
#    Set current group/peer/area/interface/main template file name.
#    Format:
#        <template type> [<protocol>=]<template file name>
#
#        If no protocol is specified, the template name is general - used
#        whenever the template for any particular protocol is undefined.
#
sub set_template
{
   local ( $template_type ) = @_;
   local ( $var_pair, $var1, $var2, $protocol, $template );

   #
   # Extract the protocol (if specified) and template name.
   &get_name( "$template_type", $_, *var_pair )
       || &error_and_die( "No template filename specified" );
   ( $var1, $var2 ) = split( "=", $var_pair );

   # If var2 is empty, assume no protocol was specified.
   if ( $var2 eq "" )
   {
      $protocol = "general";
      $template = $var1;
   }
   else
   {
      $protocol = $var1;
      $template = $var2;
   }

   # Special case:  the main template does not use the template list.
   # It has its own variable.  This overrides any value given to it
   # from the environment variable.  Any "protocol" value is ignored.
   if ( $template_type eq "main_template" )
   {
      $main_template_file = $template;
   }
   else
   {
      # Stuff the template information into the master template list.
      $g170_template_list{ $template_type, $protocol } = $template;
   }
#  print "*** $template_type template for $protocol = $template\n";
}

#
#  Subroutine:  SET_ROUTING_DOMAIN
#
#    Sets the currently active routing domain inside a sequence.
#    This routine is called both from the command processor, where the
#    domain is part of the command string, and from the mesh code,
#    which passes in the domain name as a parameter.
#
sub set_routing_domain
{
   local ( $domain_name ) = @_;
   local ( $mach_idx );

   #  Error if in global or test set scope (only valid inside a sequence).
   if (($g005_process_scope eq "GLOBAL") || ($g005_process_scope eq "TEST_SET"))
   {
      &error_and_die( "routing domain definition only valid inside a sequence" );
   }

   if ( $domain_name eq "" ) {
      &get_name( 'routing_domain', $_, *domain_name )
          || &error_and_die( "No routing domain identifier specified" ); }

   #  The domain name must be in the machine map.
   $mach_idx = &find_machine( $domain_name );

   $g011_curr_rdi = $rdi[$mach_idx];
   $g012_curr_as  = $as_number[$mach_idx];
}

#
#  Subroutine:  SET_PROTOCOL
#
#    Sets the currently active protocol (idrp, bgp, etc.).
#
sub set_protocol
{
   local ( $protocol, $supported_protocol, $supported );

   &get_name( 'protocol', $_, *protocol )
       || &error_and_die( "No protocol specified" );

   #  Check to make sure the protocol is supported by this test generator.
   $supported = $FALSE;
   foreach $supported_protocol ( @supported_protocol_list )
   {
      if ( $protocol eq $supported_protocol )
      {
         $supported = $TRUE;
         last;
      }
   }

   if ( $supported ) {
      $g006_curr_protocol = $protocol; }
   else {
      &error_and_die( "Protocol \"$protocol\" not supported by this tool" ); }
}

#
#  Subroutine:  SET_SOCKET
#
#    Establishes whether to use IDRP or IP socket to carry routes.
#
sub set_socket
{
   &get_name( 'socket', $_, *socket_name )
       || &error_and_die( "No socket type specified" );
   print "Now processing socket $socket_name ...\n";

   if ( ($socket_name ne "idrp") && ($socket_name ne "ip") ) {
      &error_and_die( "Illegal socket type \"$socket_name\"" );
   }

   $g031_proto_sock = $socket_name;
}

#
#  Subroutine:  SET_IE_TEMPLATE
#
#    Sets the import or export template file name for a test set.
#    Takes the keyword (kind of import/export) as input parameter.
#
sub set_ie_template
{
   local ( $keyword ) = @_;
   local ( $var_pair, $protocol, $template );
   local ( @protocol_list, @protocols );

   &get_name( $keyword, $_, *var_pair )
       || &error_and_die( "No protocol specified" );
   ( $protocols, $template ) = split( "=", $var_pair );
   if ( $template eq "" ) {
      &error_and_die( "No template specified" ); }

   # Find the full path of the file name.
   &find_file( *template )
      || die " *** ERROR - Cannot find template file \"$template\": $!\n";

   # Extract the list of protocols specified.
   @protocol_list = split( /[\s,]+/, $protocols );
   foreach $protocol ( @protocol_list )
   {
      KWD_SWITCH:
      {
         ( $keyword eq "import" ) && do
         {
            $g120_import_stmt_list{$import_order++, $g003_curr_test_set,
                     $protocol, SINGLE, $import_rdi_count++} = $template;
            last KWD_SWITCH;
         };

         ( $keyword eq "import_all" ) && do
         {
            $g120_import_stmt_list{$import_order++, $g003_curr_test_set,
                                   $protocol, LOOP, 0 } = $template;
            last KWD_SWITCH;
         };

         ( $keyword eq "export" ) && do
         {
            $g130_export_stmt_list{$export_order++, $g003_curr_test_set,
                     $protocol, SINGLE, $export_rdi_count++} = $template;
            last KWD_SWITCH;
         };

         ( $keyword eq "export_all" ) && do
         {
            $g130_export_stmt_list{$export_order++, $g003_curr_test_set,
                                   $protocol, LOOP, 0 } = $template;
            last KWD_SWITCH;
         };

         &error_and_die( "Unknown import/export type" );
      }     # KWD_SWITCH
   }     # foreach
}

#
#  Subroutine:  BEGIN_TEST_SET
#
sub begin_test_set
{
   $g005_process_scope = TEST_SET;

   &get_name( 'test_set', $_, *g003_curr_test_set )
       || &error_and_die( "Test set has no name" );

   push( @test_set_list, $g003_curr_test_set );
   print "Now processing test_set $g003_curr_test_set ...\n";

   #  Clear the import and export RDI list arrays and indexes.
   $import_rdi_count     = 0;
   $export_rdi_count     = 0;
}


#
#  Subroutine:  SET_VARIABLE
#  Add a variable and its value to the appropriate variable list.
#  Format: "variable=value"
#
sub set_variable
{
   local ( $var_pair, $var_name, $var_value );
   local ( $builtin, $peer );

   &get_name( 'setvar', $_, *var_pair )
       || &error_and_die( "Empty setvar statement" );

   ( $var_name, $var_value ) = split( "=", $var_pair );

   #  Allow a null value, but not a null variable name.
   if ( $var_name eq "" ) {
      &error_and_die( "No variable name for \"$var_pair\"" ); }
#  if ( $var_value eq "" ) {
#     &error_and_die( "No variable value for \"$var_pair\"" ); }

   #
   #  Check for any built-in functions.  Substitute if found.
   #
   foreach $builtin ( @builtin_list )
   {
      while ( $var_value =~ /$builtin\(.*\)/ ) {
         &subs_builtin($var_value, $builtin); }
   }

   SCOPE_SWITCH:
   {
      ( $g005_process_scope eq "GLOBAL" ) && do
      {
         $global_var_list{$var_name} = $var_value;
#        print "set_variable: Global variable $var_name set to $var_value\n";
         last SCOPE_SWITCH;
      };

      ( $g005_process_scope eq "SEQUENCE" ) && do
      {
         $sequence_var_list{$g002_curr_sequence, $var_name} = $var_value;
#        print "set_variable: Variable $var_name set to $var_value";
#        print " for sequence $g002_curr_sequence\n";
         last SCOPE_SWITCH;
      };

      ( $g005_process_scope eq "MACHINE" ) && do
      {
         $machine_var_list{$g002_curr_sequence, $g001_mach_idx, $var_name} = $var_value;
#        print "set_variable: Variable $var_name set to $var_value";
#        print " for machine $machine[$g001_mach_idx] ";
#        print " in sequence $g002_curr_sequence\n";
         last SCOPE_SWITCH;
      };

      ( $g005_process_scope eq "PEER" ) && do
      {
         foreach $peer ( @g004_curr_peers )
         {
            $peer_var_list{$g002_curr_sequence, $g001_mach_idx, $peer, $var_name} = $var_value;
#           print "set_variable: Variable $var_name set to $var_value";
#           print " for peer/interface $peer ";
#           print " of machine $machine[$g001_mach_idx] ";
#           print " in sequence $g002_curr_sequence\n";
         }
         last SCOPE_SWITCH;
      };

      ( $g005_process_scope eq "TEST_SET" ) && do
      {
         $test_set_var_list{$g003_curr_test_set, $var_name} = $var_value;
#        print "set_variable: Variable $var_name set to $var_value";
#        print " for test set $g003_curr_test_set\n";
         last SCOPE_SWITCH;
      };

      &error_and_die( "Unknown process scope" );
   }
}

#
#  Subroutine:  SUBS_BUILTIN
#     Substitute information about a machine in a variable string.
#
sub subs_builtin
{
   local ( $str, $builtin ) = @_;   # string to substitute in
   local ( $pattern, $mach_idx );

   #  The pattern we are looking for is "&<builtin>(machine)"
   #  without worrying about whitespace in between the ().
   #  Putting () around the \\S expression puts the machine name in $+.
   $pattern = "${builtin}\\(\\s*(\\S*)\\s*\\)";
   if ( $str =~ /$pattern/ )
   {
      #  If no machine is specified - i.e. &ip() - use current machine.
      $mach_idx =  ( $+ eq "" ) ? $g001_mach_idx : &find_machine( $+ );

      TYPE_SWITCH:
      {
         ( $builtin eq "&as" ) && do
         {
            $str =~ s/$pattern/$as_number[$mach_idx]/;
            last TYPE_SWITCH;
         };

         ( $builtin eq "&rdi" ) && do
         {
            $str =~ s/$pattern/$rdi[$mach_idx]/;
            last TYPE_SWITCH;
         };

         ( $builtin eq "&net" ) && do
         {
            $str =~ s/$pattern/$osi_address[$mach_idx]/;
            last TYPE_SWITCH;
         };

         ( $builtin eq "&ip" ) && do
         {
            $str =~ s/$pattern/$ip_address[$mach_idx]/;
            last TYPE_SWITCH;
         };

         &error_and_die( "Unknown builtin - \"$builtin\"" );
      }
      @_[0] = $str;
#     print "New string is \"$str\"\n";
   }
   else {
      &error_and_die( "Incorrect use of $builtin - \"$str\"" ); }
}

#
#  Subroutine:  BEGIN_SEQUENCE
#
sub begin_sequence
{
   $g005_process_scope = SEQUENCE;
   @g100_data_list     = ();
   @g110_machine_list  = ();
   %g112_area_template_list = ();

   &get_name( 'sequence', $_, *g002_curr_sequence )
       || &error_and_die( "Sequence has no name" );
   print "Now beginning sequence $g002_curr_sequence ...\n";
}

#
#  Subroutine:  END_BLOCK
#
sub end_block
{
   SCOPE_SWITCH:
   {
      ( $g005_process_scope eq "SEQUENCE" ) && do
      {
         &end_sequence;
         last SCOPE_SWITCH;
      };

      # For right now, no "end" statements for machine clauses.
      #   An "end" ends a sequence.
      ( ($g005_process_scope eq "MACHINE") ||
        ($g005_process_scope eq "PEER") ) && do
      {
         &end_sequence;
         last SCOPE_SWITCH;
      };

      ( $g005_process_scope eq "TEST_SET" ) && do
      {
         &end_test_set;
         last SCOPE_SWITCH;
      };

      &error_and_die( "Unmatched END statement" );
   }
}

#
#  Subroutine:  END_TEST_SET
#
#  Ends a test set.  Resets process scope to GLOBAL for user-defined
#  variable purposes.
#  NOTE:  an "end" statement is not required for a test set, as it is
#         for a sequence.
#
sub end_test_set
{
   print "End of test set $g003_curr_test_set ...\n";
   $g005_process_scope = GLOBAL;
}

#
#  Subroutine:  END_SEQUENCE
#
sub end_sequence
{
   local ( $test_set, $mach_idx, $machine_name, $i );
   local ( @static_routes );

   print "Ending sequence $g002_curr_sequence ...\n";
   $num_sequences++;

   #
   #  For each machine in the sequence, for each test set defined,
   #  generate a configuration file.  Restrict to a particular
   #  machine if specified in -m option.
   #
   foreach $machine_name ( @g110_machine_list )
   {
      if ( ($opt_m eq "") || ($opt_m eq $machine_name) )
      {
         foreach $test_set ( @test_set_list )
         {
            #  Generate files if no restrictions on it.
            if ( &gen_this_combo($g002_curr_sequence, $test_set) )
            {
               $mach_idx = &find_machine( $machine_name );
               @g150_default_route_list = &read_default_routes();
               @static_routes = &read_static_routes( $test_set, $mach_idx );
               &create_config_file( $mach_idx, $test_set, @static_routes );
               if ( $g201_opt_gen_res_files eq "on" )
               {
                  &create_result_file( $mach_idx, $test_set, @static_routes );
                  $num_res_files++;
               }
               $num_cfg_files++;
            }     # if &gen_this_combo
         }     # foreach $test_set
      }     # if machine matches
   }    # foreach $machine_name

   #  Reset the RDI to their original values (in machine map)
   for ( $i = 0; $i <= $#machine; $i++ )
   {
         $rdi[$i] = $save_rdi[$i];
         $as_number[$i] = $save_as[$i];
   }

   #  Reset process scope to global.
   $g005_process_scope = GLOBAL;
}

#
#  Subroutine:  GEN_THIS_COMBO
#    Tests to see if there are any restrictions on generating this
#    sequence/test set combo.  The most inclusive set is used.
#
sub gen_this_combo
{
   local ( $sequence, $test_set ) = @_;
   local ( $entry, $seq_flag, $test_flag, $gen_flag );

   # Loop through the generation list for restrictions.
   # If no entries in list, generate everything.
   $gen_flag = 1;
   foreach $entry( @g160_generation_list )
   {
      #  If at least one entry, now do not generate unless it matches.
      $gen_flag = 0;

      # Since we do the most inclusive, if "all" is present, exit at once.
      if ( $entry eq "all" )
      {
         $gen_flag = 1;
         last;
      }

      # Return "generate" if a match is found.
      ( $seq_flag, $test_flag ) = split( /[\s,]+/, $entry );

      if ( (($seq_flag  eq $sequence) || ($seq_flag  eq "*")) &&
           (($test_flag eq $test_set) || ($test_flag eq "*")) )
      {
         $gen_flag = 1;
         last;
      }
   }

#  print "** gen_this_combo:  gen_flag = $gen_flag\n";
   return( $gen_flag );
}

#
#  Subroutine:  BEGIN_MACHINE
#
sub begin_machine
{
   local ( $machine_name ) = @_;

   #  Error if not in sequence or lower scope.
   if ( ($g005_process_scope eq "GLOBAL")
     || ($g005_process_scope eq "TEST_SET") ) {
      &error_and_die( "Machines are defined only inside a sequence" ); }

   $g005_process_scope = MACHINE;

   @g004_curr_peers = ();

   #  If this routine was called with a name passed in, use that.
   #  Otherwise, extract the name from the attached argument.
   if ( $machine_name eq "" ) {
      &get_name( 'machine', $_, *machine_name )
          || &error_and_die( "Machine has no name" ); }

   $g001_mach_idx = &find_machine( $machine_name );
   if ( $g001_mach_idx == -1 )
   {
      &error_and_die( "Machine $machine_name not found in table" );
   }

   #  Stop if no routing domain has been specified.
   if ( $g011_curr_rdi eq "" ) {
      &error_and_die( "No routing domain specified" ); }

   print "Now processing machine $machine_name ...\n";

   &push_unique( $machine_name, *g110_machine_list );

   # Replace the machine's default RDI and AS number (from machine map)
   # with the currently active RDI and AS number.
   $rdi[$g001_mach_idx] = $g011_curr_rdi;
   $as_number[$g001_mach_idx] = $g012_curr_as;
}


#
#  Subroutine:  SET_RIB
#  Assign a "rib" statement at global or machine scope.
#  Format of rib statement:  rib <rib_id>=<template>
#
sub set_rib
{
    local( $sequence, $mach_idx );
    local( $name_str, $rib_number, $template );
 
    &get_name( 'rib', $_, *name_str )
        || &error_and_die( "No value in \"rib\" statement" );
 
    ( $rib_number, $template ) = split( "=", $name_str );
 
    if ( $rib_number eq "" ) {
       &error_and_die( "No RIB ID in \"$name_str\"" ); }
    if ( $template eq "" ) {
       $template = $g170_template_list{rib_template, $g006_curr_protocol}; }
    if ( $template eq "" ) {
       $template = $g170_template_list{rib_template, general}; }
    if ( $template eq "" ) {
          &error_and_die( "No RIB template specified" ); }

    # Find the full path of the file name.
    &find_file( *template )
       || die " *** ERROR - Cannot find template file \"$template\": $!\n";

    SCOPE_SWITCH:
    {
       ( $g005_process_scope eq "GLOBAL" ) && do
       {
          $sequence = $ALL;
          $mach_idx = $ALL;
          last SCOPE_SWITCH;
       };
 
       ( $g005_process_scope eq "SEQUENCE" ) && do
       {
          $sequence = $g002_curr_sequence;
          $mach_idx = $ALL;
          last SCOPE_SWITCH;
       };
 
       ( $g005_process_scope eq "MACHINE" ) && do
       {
          $sequence = $g002_curr_sequence;
          $mach_idx = $g001_mach_idx;
          last SCOPE_SWITCH;
       };
 
       &error_and_die( "Invalid scope for declaring a RIB statement" );
    }
 
    $entire_ribs_list{ $sequence, $mach_idx, $rib_number } = $template;
}

#
#  Subroutine:  SET_RIBS_SUPPORTED
#  Assign a "ribs-supported" clause in a peer statement.
#
sub set_ribs_supported
{
   local ( $peer, $ribs );

   &get_name( 'ribs-supported', $_, *ribs )
       || &error_and_die( "No value in \"ribs-supported\" statement" );

   #  Error if not currently processing a peer.
   if ( $g005_process_scope ne "PEER" ) {
      &error_and_die( "No peer defined for \"ribs-supported\" statement" ); }

   foreach $peer ( @g004_curr_peers )
   {
      $peer_ribs_list{$g002_curr_sequence, $g001_mach_idx, $peer} = $ribs;
#     print "Supporting ribs \"$ribs\" for peer $peer\n";
   }
}

#
#  Subroutine: SET_FULL_MESH
#    Define a list of machines in the current sequence and set them
#    to peer with each other in a full mesh.
#    Format: mesh <peer type>=<machine list>
#
sub set_full_mesh
{
   local ( @machine_list, $machines, $machine_name );
   local ( @peer_list, $peer, $peer_type );
   local ( $var_pair );

   #  Error if not in sequence or lower scope.
   if ( ($g005_process_scope eq "GLOBAL")
     || ($g005_process_scope eq "TEST_SET") ) {
      &error_and_die( "Machines are defined only inside a sequence" ); }

   &get_name( 'mesh', $_, *var_pair )
       || &error_and_die( "No peer type or list of machines specified" );

   ( $peer_type, $machines ) = split( "=", $var_pair );

   #  If no equal sign, one of the two required fields is missing.
   if ( $machines eq "" ) {
       &error_and_die( "Peer type or machine list missing in mesh statement" ); }

   @machine_list = split( /[\s,]+/, $machines );

   #  Define a full peer relationship among the machines in the mesh list.
   foreach $machine_name ( @machine_list )
   {
      #  If this is an external mesh, set the current routing domain/as
      #  to the RDI/AS of the current machine.  For an internal mesh, the
      #  routing domain should already be specified by an earlier command,
      #  so print an error if not defined.
      if ( $peer_type eq "external" ) {
         set_routing_domain( $machine_name ); }
      elsif ( $g011_curr_rdi eq "" ) {
         &error_and_die( "No routing domain specified for $peer_type mesh" ); }

      #  Register this as the current machine.
      &begin_machine( $machine_name );

      # The peer list is every machine in the machine list except the one
      # now being processed.
      @peer_list = ();
      foreach $peer ( @machine_list )
      {
         if ( $peer ne $machine_name ) {
           push( @peer_list, $peer ); }
      }
      &set_peers( $peer_type, @peer_list );
   }
}

#
#  Subroutine: SET_PEERS
#    Assign a list of peers to the current machine in a sequence.
#    Peer type (internal, external, ...) passed in as parameter.
#
sub set_peers
{
   local ( $peer_type, @peer_list ) = @_;
   local ( $machines, $peer, $template, $new_entry );

   $g005_process_scope = PEER;

   if ( @peer_list == () ) {
      &get_name( $peer_type, $_, *machines )
          || &error_and_die( "No peers listed" ); }

   # Get a template name if one is specified.  Otherwise set to current.
   ( $machines, $template ) = split( "=", $machines );
   if ( $template eq "" ) {
      $template = $g170_template_list{peer_template, $g006_curr_protocol}; }
   if ( $template eq "" ) {
      $template = $g170_template_list{peer_template, general}; }
   if ( $template eq "" ) {
      die "***** ERROR -- No peer template defined\n"; }

   # Find the full path of the file name.
   &find_file( *template )
      || die " *** ERROR - Cannot find template file \"$template\": $!\n";

   # Extract the peer names.
   if ( @peer_list == () ) {
      @g004_curr_peers = split( /[\s,]+/, $machines ); }
   else {
      @g004_curr_peers = @peer_list; }

   foreach $peer ( @g004_curr_peers )
   {
      print "Processing $peer_type peer $peer ...\n";
#     print "*** Protocol = $g006_curr_protocol\n";

      # Data entry = <protocol> <machine> <peer type> <peer> <template>
      $new_entry = join( "$;", $g006_curr_protocol, $machine[$g001_mach_idx],
                               $peer_type, $peer, $template );
      push( @g100_data_list, $new_entry );
   }
}

#
#  Subroutine: ADD_INTERFACES
#    Add a list of OSPF or RIP interfaces to the data list for this sequence.
#
sub add_interfaces
{
   local ( $machines, $interface, $template, $new_entry );

   $g005_process_scope = PEER;

   &get_name( "interface", $_, *machines )
       || &error_and_die( "No interfaces listed" );

   # Get a template name if one is specified.  Otherwise set to current.
   ( $machines, $template ) = split( "=", $machines );
   if ( $template eq "" ) {
      $template = $g170_template_list{intf_template, $g006_curr_protocol}; }
   if ( $template eq "" ) {
      $template = $g170_template_list{intf_template, general}; }
   if ( $template eq "" ) {
      die "***** ERROR -- No interface template defined\n"; }

   # Find the full path of the file name.
   &find_file( *template )
      || die " *** ERROR - Cannot find template file \"$template\": $!\n";

   # Extract the interface names.  Assign to g004_curr_peers for processing
   # user variables at peer/interface scope.
   @g004_curr_peers = split( /[\s,]+/, $machines );

   foreach $interface ( @g004_curr_peers )
   {
      print "Processing interface $interface ...\n";

      # Data entry = <protocol> <machine> <area> <interface> <template>
      $new_entry = join( "$;", $g006_curr_protocol, $machine[$g001_mach_idx],
                               $g007_curr_area, $interface, $template );
      push( @g100_data_list, $new_entry );
   }
}

#
#  Subroutine:  SET_ROUTE_SET
#  Determine the static routes to be used for a given test set.
#
#  Format:
#     route_set <machine name> | "default" | "all" = <route set ID>
#
sub set_route_set
{
   local ( $var_pair, $param_1, $param_2, $mach_idx, $route_set_id );

   #  Error if not in test set scope.
   if ( $g005_process_scope ne "TEST_SET" ) {
      &error_and_die( "route_set kwd only usable inside a test set" ); }

   &get_name( 'route_set', $_, *var_pair )
       || &error_and_die( "No route set name specified" );

   #  Get the machine name and route set ID.
   #  If only one parameter specified, assign as route set ID and assume
   #  "all" as machine name.
   ( $param_1, $param_2 ) = split( "=", $var_pair );
   if ( $param_2 eq "" )
   {
      $machine = "all";
      $route_set_id = $param_1;
   }
   else
   {
      $machine = $param_1;
      $route_set_id = $param_2;
   }

   #  Store the machine/route set ID pair in the global list.  Looking up
   #  the machine index will verify that this is a valid machine.
   if    ( $machine eq "all" )     { $mach_idx = "all" }
   elsif ( $machine eq "default" ) { $mach_idx = "default" }
   else  { $mach_idx = &find_machine($machine) };

   $g140_route_set_id_list{$g003_curr_test_set, $mach_idx} = $route_set_id;
#  print "** route_set_id_list{$g003_curr_test_set, $mach_idx} = $route_set_id\n";
}

#
#  Subroutine:  FIND_FILE
#  Try to find a file name, returning directory/filename in parameter
#  and success/fail flag as return value.
#
#  If the path is not coded in, use the following search path:
#     1.  Current directory.
#     2.  Test subdirectory with the name of the current test.
#     3.  Template directory.
#
sub find_file
{
   local ( *filename ) = @_;

   # If path is specified, just return.
   if ( $filename =~ /\// ) {
      return( 1 ); }

   foreach $try ( "./$filename",
                  "$gated_test_dir/$g021_curr_test_name/$filename",
                  "$gated_template_dir/$filename" )
   {
      if ( -e $try )
      {
         $filename = $try;
         return( 1 );
      }
#     else { print "*** find_file: no match found for \"$try\"\n"; }
   }

   # If we got here, no match was found.
   return( 0 );
}

#
#  Subroutine:  FIND_MACHINE
#  Find the machine's index in the table.
#
sub find_machine
{
   local ( $mach_name ) = @_[0];

#  print "Looking for machine '$mach_name'\n";

   for ( $i = 0; $i <= $#machine; $i++ )
   {
      if ( $machine[$i] eq $mach_name )
      {
         return $i;
      }
   }

   #
   #  Since we got here, the search failed.
   #
   &error_and_die( "Machine \"$mach_name\" not in machine map." );
}

#
#  Subroutine:  PUSH_UNIQUE
#  Push an entry onto a list, but only if not in there already.
#
sub push_unique
{
   local ( $new_entry, *list ) = @_;
   local ( $in_list ) = 0;

   foreach $list_entry ( @list )
   {
      if ( $list_entry eq $new_entry )
      {
         $in_list = 1;
         last;
      }
   }

   if ( ! $in_list )
   {
      push( @list, $new_entry );
   }
}

#
#  Subroutine:  GET_NAME
#  Get the name of a keyword (sequence, machine, etc.)
#
sub get_name
{
   local ( $keyword, $line, *name ) = @_;
   $pattern = "^\\s\*$keyword\\s\*";

#  print "Keyword = '$keyword'\n";
#  print "Line    = '$line'\n";
#  print "Pattern = '$pattern'\n";

   ( $dummy, $name ) = split( /$pattern/, $line );
   if ( $name =~ /\n$/ )
   {
      chop( $name );
   }

   # Remove trailing whitespace.
   while ( $name =~ /\s$/ )
   {
      chop( $name );
   }

#  print "Returned name = '$name'\n\n";

   return( $name eq "" ? 0 : 1 );
}

#
#  Subroutine:  PRINT_OVERVIEW
#     Print the overview message for this program.
#
sub print_overview
{
   print "\n$BASENAME:\n";
   print "     This program generates gated configuration (gated.conf) files and\n";
   print "     expected result (expected.txt) files for every machine in a network\n";
   print "     by interpreting a simplified language used to describe the network\n";
   print "     configurations, static routes, and import/export policy.\n";
   print "     Version: $VERSION ($VDATE)\n";
}

#
#  Subroutine:  PRINT_USAGE
#     Print the usage message for this program.
#
sub print_usage
{
   print "\nUsage: $BASENAME [-EhILW] [-m machine_name] [-t test_name] <test config file>\n";
   print "Options:\n";
   print "   -m <machine_name> - Generate files only for that machine\n";
   print "   -t <test_name>    - Go to this test directory and make files\n";
   print "   -E                - Do not generate expected results files\n";
   print "   -h                - Print this usage and overview message\n";
   print "   -I                - Do not generate informational messages (do print errors)\n";
   print "   -L                - Do not generate export loop code\n";
   print "   -W                - Do not generate warning messages (do print errors)\n";
   print "\n";
}

#
#  Subroutine:  ERROR_AND_DIE
#     Print an error message and halt program.
#
sub error_and_die
{
   local ( $error_message ) = @_;

   print " *** ERROR - ";
   print "$error_message\n";
   die "      At line # $line_number of file $config_file.\n";
}

#
# end of file
